#include <stdio.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#define UP(semid, semnum)     {struct sembuf up = {semnum, 1, 0};    \
                               semop(semid, &up, 1); }
#define DOWN(semid, semnum)  {struct sembuf down = {semnum, -1, 0};  \
                               semop(semid, &down, 1); }
#define INIT(semid, semnum)   UP(semid, semnum);

#define NPROCS 5
#define A 0
#define B 10
#define N 20
#define H (((float)B - (float)A) / N)
int mutex;
float *soma;
void integral(int i);
float f(float t);
int init_sems(int nsems);
void free_sems(int semid);

main() {
  int i, shmid, status;

  // alocar e mapear um segmento de memria partilhada para a soma
  shmid = shmget(IPC_PRIVATE, sizeof(float), S_IRUSR | S_IWUSR);
  soma = (float *) shmat(shmid, NULL, 0);
  // alocar um mutex
  mutex = init_sems(1);

  *soma = 0;
  for (i = 0; i < NPROCS; i++) 
    if (fork() == 0) {
      integral(i);
      exit(0);
    }

  // pai espera que todos os filhos terminem
  for (i = 0; i < NPROCS; i++)
    wait(&status);

  printf("Area total= %f\n", H * (*soma + (f(B) - f(A)) / 2));

  // remover o mutex
  free_sems(mutex);
  // libertar e remover o segmento de memria partilhada
  shmdt(soma);
  shmctl(shmid, IPC_RMID, NULL);
}

void integral(int region) {
  float t, soma_parcial;
  int i;

  t = A + region * (B - A) / NPROCS;  // incio do intervalo
  soma_parcial = 0;
  for (i = 0; i < N / NPROCS; i++) {
    soma_parcial += f(t);
    t += H;
  }
  DOWN(mutex, 1);
  *soma = *soma + soma_parcial;
  UP(mutex, 1);
}

float f(float t) {
  return t;
}

int init_sems(int nsems) {
  int semid, i;
  semid = semget(IPC_PRIVATE, nsems, S_IRUSR | S_IWUSR);
  for (i = 0; i < nsems; i ++)
    INIT(semid, i);  
  return semid;
}

void free_sems(int semid) {
  semctl(semid, 0, IPC_RMID);
}
